;;; $Id: tc-subst.scm,v 1.53 2006/02/05 15:03:19 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; Generalized Substitutions and Unification
;;; 
;;; AUTHOR: Mark P. Jones as generalized by Gary T. Leavens
;;;         and translated to Scheme by Brian Dorn
;;;         (The original is in the Prolog demo for Hugs)

(module tc-subst (lib "typedscm.ss" "typedscm")

(provide tc:make-subst tc:subst? tc:subst-compose
	 tc:subst-apply tc:make-null-subst
         tc:unifiable-md tc:unifiable-md? tc:unifiable-md-dict
	 tc:make-unifiable-md
         tc:mgu tc:subtype-mgu tc:vars-in
         tc:list-unify tc:list-subtype-unify tc:rename-vars
         ;; all-from tc-logical-variables.scm
         tc:logicalvar? tc:variable? tc:make-logicalvar
         tc:symbol->logicalvar tc:new-logicalvar tc:new-lv-list
         tc:logicalvar->name tc:logicalvar->level tc:variable-eq?
	 tc:logicalvar->uniq-sym
         ;; all-from tc-attrib-pair.scm
         tc:attrib-pair-of : tc:attrib-pair->syn tc:attrib-pair->type
         tc:tree-of tc:root tc:leaves tc:node 
         tc:print-map tc:tree-print
	 )
  
(deftype tc:make-subst
  (forall (Term) (-> (tc:logicalvar Term (-> (tc:logicalvar) Term)) tc:subst)))
(deftype tc:subst?
  (type-predicate-for tc:subst))
(deftype tc:subst-compose
  (-> (tc:subst tc:subst) tc:subst))
(deftype tc:mgu
  (forall (Term)
    (-> (tc:unifiable-md) (-> (Term Term) (list-of tc:subst)))))
(deftype tc:subtype-mgu
  (forall (Term)
    (-> (tc:unifiable-md) (-> (Term Term) (list-of tc:subst)))))
(deftype tc:vars-in
  (forall (Term)
    (-> (tc:unifiable-md)
	(-> (Term) (list-of tc:logicalvar)))))
(deftype tc:list-unify
  (forall (Term)
    (-> (tc:unifiable-md)
	(-> ((list-of Term) (list-of Term)) (list-of tc:subst)))))
(deftype tc:list-subtype-unify
  (forall (Term)
    (-> (tc:unifiable-md)
	(-> ((list-of Term) (list-of Term)) (list-of tc:subst)))))
(deftype tc:rename-vars
  (forall (Term)
    (-> (tc:unifiable-md)
	(-> (number)
	    (-> (Term) Term)))))
(deftype tc:subst-apply
  (forall (Term) (-> (tc:subst tc:logicalvar) Term)))

(require (lib "maybe.scm" "typedscm")
         (lib "tc-logical-variables.scm" "typedscm")
         (lib "tc-attrib-pair.scm" "typedscm")
	 (lib "tc-env.scm" "typedscm"))

;; We don't have polymorphic datatype definitions, so we use this instead...
(deftype Term? (type-predicate-for datum))
(define Term? datum?)

;;           SUBSTITUTIONS

;; A substitution is a function from logical variables to terms.

(define-datatype tc:subst tc:subst?
  (tc:normal-subst
   (var-te-maker (-> (tc:logicalvar?) Term?))
   (env (tc:env-of symbol? Term?)))
  (tc:all-subst
   (var-te-maker (-> (tc:logicalvar?) Term?))
   (new-lev number?)))

(define tc:subst->env
  (lambda (s)
    (cases tc:subst s
	   (tc:normal-subst (var-te-maker env) env)
	   (else (error "Cannot extract env from all subst")))))

(define tc:make-null-subst
  (lambda (var-te-maker)
    (tc:normal-subst var-te-maker (tc:env-empty))))

(define tc:make-subst
  (lambda (lv t var-te-maker)
    (tc:normal-subst var-te-maker
		     (tc:env-update
		      (tc:logicalvar->uniq-sym lv)
		      t
		      (tc:env-empty)))))

(define tc:subst-apply
  (lambda (s lv)
;;;    (displayln "SUBST APPLY: ")
;;;    (displayln "S: " s)
;;;    (displayln "LV: " lv)
    (cases tc:subst s
	   (tc:normal-subst
	    (var-te-maker env)
	    (let ((lv-sym (tc:logicalvar->uniq-sym lv)))
	      (if (tc:env-bound? lv-sym env)
		  (tc:env-value lv-sym env)
		  (var-te-maker lv))))
	   (tc:all-subst
	    (var-te-maker lev)
	    (var-te-maker
	     (tc:make-logicalvar
	      (tc:logicalvar->name lv)
	      lev))))))
	    

;;          SUBSTITUTION CREATION AND APPLICATION

;; Terms are defined by the user of this module.
;; Thus we assume the following capabilities when needed
;; to create and apply substititions to terms.

;; nullSubst : tc:subst
;; nullSubst is the empty substitution;
;;            it is like the identity function in that, for all lv,
;; 	              (nullSubst lv) = (toTerm lv),
;;            where toTerm is a function that takes a logical variable
;;                  and makes a term out of it. Typically toTerm will be some
;;                  constructor in the grammar for terms.
;;
;; bind : (-> (tc:logicalvar Term) tc:subst)
;; (bind lv t) is the substitution that maps the Tc:Logicalvar lv to the term t,
;;            but otherwise behaves like nullSubst.
;;
;; app : (-> (tc:subst) (-> (Term) Term))
;; app s     extends the substitution s to a function mapping terms to terms
;; 
;;  We require that the following equations hold:
;;    ((app nullSubst) t) = t
;;    (app (bind lv t) (nullSubst lv'))
;;         = (if (equal? lv' lv) t (nullSubst lv'))

;; Using these operations one can define composition of substitutions
;; as follows.

(define tc:subst-compose
  (lambda (s1 s2)
    (cases tc:subst s1
	   (tc:normal-subst
	    (var-te-maker env)
	    (tc:normal-subst
	     var-te-maker
	     (tc:env-join env (tc:subst->env s2))))
	   (else (error "Cannot compose all subst!")))))
      
;; Composition of substitutions should be is a monoid homomorphism
;;    from (tc:subst,               nullSubst,       tc:subst-compose)
;;    to   ((-> (Term) Term),       (lambda (t) t),  compose)
;; in the sense that the following equations hold.
;;       (app (tc:subst-compose s1 s2)) = (compose (app s1) (app s2))
;;       (tc:subst-compose s nullSubst) = s
;;       (tc:subst-compose nullSubst s) = s

;;           UNIFICATION

;; To be unifiable, terms must have these additional capabilities.

;; to-term : (-> (tc:logicalvar) Term)
;; (to-term lv) is the logical variable lv made into a term

;; get-var : (-> (Term) (maybe-of tc:logicalvar))
;; (get-var t)  is (make-something lv) when t is a term representing
;;               a logical variable lv, and (make-nothing) when t
;;               does not represent a logical variable

;; subterms : (-> (Term) (list-of Term))
;; (subterms t) is a list of the subterms of t

;; same-kind: (-> (Term Term) boolean)
;; (same-kind t1 t2) is True when t1 and t2 are the same kind of term,
;;                   that is they have the same operator

;;    We require that the following equations hold:
;;       (getVar (to-Term lv)) = (make-something lv)
;;       (subterms (to-Term lv)) = '()
;;       (equal? (getVar t1) (make-nothing)) =
;;               (equal? (getVar t2) (make-nothing)), if (same-kind t1 t2)


;;         SUBTYPE UNIFICATION

;; In addition, the following capabilities must be present to
;; allow for unification of subtypes in terms.

;; contravar-subterms : (-> (Term) (list-of Term))
;; (contravar-subterms t) is a list of the subterms of t which
;; must be treated as contravariant in subtyping

;; covar-subterms : (-> (Term) (list-of Term))
;; (covar-subterms t) is a list of the subterms of t which
;; must be treated as covariant in subtyping

;; invar-subterms : (-> (Term) (list-of Term))
;; (invar-subterms t) is a list of the subterms of t which
;; must be treated as invariant in subtyping

;; subtype-replace : (-> (Term Term) (pair-of Term Term))
;; (subtype-replace subt supert) returns a pair like (subt . supert)
;; except any special replacements due to subtyping have been made
;;  For example, expansion of varargs function types

;; intersection-type? : (-> (Term) boolean)
;; (intersectiontype? t) returns #t iff t is an intersection type

;; find-intersection-subtyping : (-> (Term Term) (list-of (subst-of Term)))
;; (find-intersection-subtyping intersection sought) performs subtype unification
;; and returns the result substitution of the first type in intersection that
;; unifies with sought

;; find-intersection-supertyping : (-> (Term Term) (list-of (subst-of Term)))
;; (find-intersection-subtyping sought intersection) attempts to find a
;; substitution that makes "sought" a subtype of *each* type in "conj-type".
;; Returns the empty list if no such environment exists.

;; is-bottom? : (-> (Term) boolean)
;; (is-bottom? t) returns #t iff t is the bottom type

;; is-top? : (-> (Term) boolean)
;; (is-top? t) returns #t iff t is the top type

;; is-error? : (-> (Term) boolean)
;; (is-error? t) returns #t iff t is an error type

;;         ERROR GENERATION

;; gen-error-mismatch : (-> (datum Term Term) Term)
;; (gen-error-mismatch syn expected inferred) returns an error type
;; expression that contains an appropriate error message
;; generated from the expected and inferred types.
;; It is called when there is a unification error.

;; gen-error-rule : (-> (datum (list-of (tc:atree-of Term))) Term)
;; (gen-error-rule syn trees) returns an error type expression
;; that is appropriate for when the rule for syntax syn fails

(define-datatype tc:unifiable-md tc:unifiable-md?
  (tc:unifiable-md-dict
   (to-term (-> (tc:logicalvar?) Term?))
   (get-var (-> (Term?) (maybe-of tc:logicalvar?)))
   (subterms (-> (Term?) (list-of Term?)))
   (same-kind (-> (Term? Term?) boolean?))
   (nullSubst tc:subst?)
   (bind (-> (tc:logicalvar? Term?) tc:subst?))
   (app (-> (tc:subst?) (-> (Term?) Term?)))
   ;; For Subtyping
   (contravar-subterms (-> (Term?) (list-of Term?)))
   (covar-subterms (-> (Term?) (list-of Term?)))
   (invar-subterms (-> (Term?) (list-of Term?)))
   (subtype-replace (-> (Term? Term?) (pair-of Term? Term?)))
   (intersection-type? (-> (Term?) boolean?))
   (find-intersection-subtyping (-> (Term? Term?) (list-of tc:subst?)))
   (find-intersection-supertyping (-> (Term? Term?) (list-of tc:subst?)))
   (is-bottom? (-> (Term?) boolean?))
   (is-top? (-> (Term?) boolean?))
   (is-error? (-> (Term?) boolean?))
   ;; For Error Generation
   (gen-error-mismatch (-> (datum? Term? Term?) Term?))
   (gen-error-rule (-> (datum? (list-of (tc:tree-of Term?))) Term?))
   ;; For dealing with specific type exprs
   (is-dec-type-expr? (-> (Term?) boolean?))
   (dec-type-expr->env (-> (Term?) datum?))  ;;[[FIXME: need envs]]
   ))

(define tc:make-unifiable-md
  (has-type-trusted
    (forall (Term Env)
      (-> (;; to-term
	   (-> (tc:logicalvar) Term)
	   ;; get-var
	   (-> (Term) (maybe-of tc:logicalvar))
	   ;; subterms
	   (-> (Term) (list-of Term))
	   ;; same-kind
	   (-> (Term Term) boolean)
	   ;; nullSubst
	   tc:subst
	   ;; bind
	   (-> (tc:logicalvar Term) tc:subst)
	   ;; app
	   (-> (tc:subst) (-> (Term) Term))
	   ;; For Subtyping
	   ;; contravar-subterms
	   (-> (Term) (list-of Term))
	   ;; covar-subterms
	   (-> (Term) (list-of Term))
	   ;; invar-subterms
	   (-> (Term) (list-of Term))
	   ;; subtype-replace
	   (-> (Term Term) (pair-of Term Term))
	   ;; intersection-type
	   (-> (Term) boolean)
	   ;; find-intersection-subtyping
	   (-> (Term Term) (list-of tc:subst))
	   ;; find-intersection-supertyping
	   (-> (Term Term) (list-of tc:subst))
	   ;; is-bottom
	   (-> (Term) boolean)
	   ;; is-top
	   (-> (Term) boolean)
	   ;; is-error
	   (-> (Term) boolean)
	   ;; For Error Generation
	   ;; gen-error-mismatch
	   (-> (datum Term Term) Term)
	   ;; gen-error-rule
	   (-> (datum (list-of (tc:tree-of Term))) Term)
	   ;; For dealing with specific type exprs
	   ;; is-dec-type-expr
	   (-> (Term) boolean)
	   ;; dec-type-expr->env
	   (-> (Term) Env)
	   )
	  tc:unifiable-md))
    (lambda (to-term
	     get-var subterms same-kind nullSubst bind app
	     contravar-subterms covar-subterms invar-subterms
	     subtype-replace intersection-type?
	     find-intersection-subtyping find-intersection-supertyping
	     is-bottom? is-top? is-error?
	     gen-error-mismatch gen-error-rule
	     is-dec-type-expr? dec-type-expr->env)
      (tc:unifiable-md-dict
       to-term get-var subterms same-kind nullSubst bind app
       ;; For Subtyping
       contravar-subterms covar-subterms invar-subterms
       subtype-replace intersection-type?
       find-intersection-subtyping find-intersection-supertyping
       is-bottom? is-top? is-error?
       ;; For Error Generation
       gen-error-mismatch gen-error-rule
       ;; For dealing with specific type exprs
       is-dec-type-expr? dec-type-expr->env))))

;; The function mgu finds the most general unifier of two terms, if it exists
;; It returns its result in a list; if no unifier exists, then () is returned.

(define tc:mgu
  (lambda (md)
    (cases tc:unifiable-md md
       (tc:unifiable-md-dict
	(to-term get-var subterms same-kind nullSubst bind app
		 contravar-subterms covar-subterms invar-subterms
		 subtype-replace intersection-type? find-intersection-subtyping find-intersection-supertyping
		 is-bottom? is-top? is-error? gen-error-mismatch gen-error-rule
		 is-dec-type-expr? dec-type-expr->env)
	(let ((vars-in (tc:vars-in md)))
	  (lambda (t1 t2)
	    (maybe-test
	     (get-var t1)
	     (lambda (x) (maybe-test
			  (get-var t2)
			  (lambda (y)
			    (list 
			     (if (equal? x y)
				 nullSubst
				 (bind x (to-term y)))))
			  (lambda ()
			    (if (not (member x (vars-in t2)))
				(list (bind x t2))
				'()))))
	     (lambda () (maybe-test
			 (get-var t2)
			 (lambda (y)
			   (if (not (member y (vars-in t1)))
			       (list (bind y t1))
			       '()))
			 (lambda ()
			   (if (same-kind t1 t2)
				((tc:list-unify md) (subterms t1) (subterms t2))
				'())))))))))))

(define tc:subtype-mgu
  (lambda (md)
    (cases tc:unifiable-md md
	   (tc:unifiable-md-dict
	    (to-term get-var subterms same-kind nullSubst bind app
		     contravar-subterms covar-subterms invar-subterms
		     subtype-replace intersection-type? find-intersection-subtyping find-intersection-supertyping
		     is-bottom? is-top? is-error? gen-error-mismatch gen-error-rule
		     is-dec-type-expr? dec-type-expr->env)
	(let ((vars-in (tc:vars-in md)))
	  (lambda (subtype supertype)
	    ;; Deal with replacement if needed
	    (let* ((replacement-res (subtype-replace subtype supertype))
		   (subtype (car replacement-res))
		   (supertype (cdr replacement-res)))
;;	      (displayln "REPLACE RES SUB " subtype)
;;	      (displayln "REPLACE RES SUP " supertype)
	      ;; Deal with intersection types here
	      (cond
	       ((intersection-type? subtype) (find-intersection-subtyping subtype supertype))
	       ((intersection-type? supertype) (find-intersection-supertyping subtype supertype))
	       (else
		;; All other cases we just move ahead like normal
		(maybe-test
		 (get-var subtype)
		 (lambda (subtype-var)	; (subtype-unify-variable)
		   (maybe-test
		    (get-var supertype)  
		    (lambda (supertype-var)
		      (list ; subtype is a variable and supertype is a variable
		       (if (equal? subtype-var supertype-var)
			   nullSubst
			   (bind subtype-var (to-term supertype-var)))))   ;;[[[DOES ORDER OF BINDING MATTER HERE??]]]
			   ;;(bind supertype-var (to-term subtype-var)))))  
		    (lambda ()
		      ;; subtype is a variable and supertype is not
		      (cond
		       ;; occurs check
		       ((member subtype-var (vars-in supertype)) '()) ; failure
		       ;; supertype is poof, subtype must also be poof
		       ((is-bottom? supertype) (list (bind subtype-var supertype)))
		       ;; supertype is datum, leave subtype alone
		       ((is-top? supertype) (list nullSubst))
		       ;; supertype is something non-special, just bind
		       (else (list (bind subtype-var supertype)))))))
		 (lambda ()
		   (maybe-test
		    (get-var supertype)
		    (lambda (supertype-var)
		      ;; subtype is not a variable and supertype is (supertype-unify-variable)
		      (cond
		       ;; occurs check
		       ((member supertype-var (vars-in subtype)) '()) ; failure
		       ;; Subtype is poof, leave super alone
		       ((is-bottom? subtype) (list nullSubst))
		       ;; Subtype is datum, bind it (can combine with below)
		       ((is-top? subtype) (list (bind supertype-var subtype)))
		       ;; Otherwise, just bind
		       (else (list (bind supertype-var subtype)))))
		    (lambda ()
		      ;; neither are variables
		      (cond
		       ((and (is-bottom? subtype) (not (is-error? supertype))) 
			(list nullSubst))
		       ((and (is-top? supertype) (not (is-error? supertype)))
			(list nullSubst))  
		       ((same-kind subtype supertype)
			(let ((list-sub-unify (tc:list-subtype-unify md))
			      (list-unify (tc:list-unify md)))
			  ;; Try unifying contravariant part
			  (let ((contra-subl        
				 (list-sub-unify
				  (contravar-subterms supertype)
				  (contravar-subterms subtype))))
			    (if (null? contra-subl)
				;; Contravariant parts do not unify
				'()
				;; Contravariant checks, try covariant
				(let ((covar-subl
				       (list-sub-unify
					(map (app (car contra-subl)) (covar-subterms subtype))
					(map (app (car contra-subl)) (covar-subterms supertype)))))
				  (if (null? covar-subl)
				      ;; Error in covariant parts
				      '()
				      ;; So far so good, try invariant parts
				      (let ((tempsub
					     (tc:subst-compose (car contra-subl) (car covar-subl) )))
					(let ((invar-subl
					       (list-unify
						(map (app tempsub) (invar-subterms subtype))
						(map (app tempsub) (invar-subterms supertype)))))
					  (if (null? invar-subl)
					      ;; Error on invariant parts
					      '()
					      ;; Everything checks, give back sub
					      (list (tc:subst-compose tempsub (car invar-subl))))))))))))		   
		       (else '()))	; unification error
		      )))))))))))))
					  
(define tc:vars-in
  (lambda (md)
    (cases tc:unifiable-md md
       (tc:unifiable-md-dict
	(to-term get-var subterms same-kind nullSubst bind app
		 contravar-subterms covar-subterms invar-subterms
		 subtype-replace intersection-type? find-intersection-subtyping find-intersection-supertyping
		 is-bottom? is-top? is-error? gen-error-mismatch gen-error-rule
		 is-dec-type-expr? dec-type-expr->env)
	 (letrec ((recurse
	   (lambda (t)
      	     (maybe-test (get-var t)
			 (lambda (lv) (list lv))
			 (lambda ()
			   (remove-duplicates
			    (apply append (map recurse (subterms t)))))))))
	   recurse)))))

(deftype tc:list-unify-maker
  (forall (Term)
    (-> ((-> (tc:unifiable-md) (-> (Term Term) (list-of tc:subst)))
	 tc:unifiable-md)
	(-> ((list-of Term) (list-of Term)) (list-of tc:subst)))))
(define tc:list-unify-maker
  (lambda (mgu-proc md)
    (cases tc:unifiable-md md
	   (tc:unifiable-md-dict
	    (to-term get-var subterms same-kind nullSubst bind app
		     contravar-subterms covar-subterms invar-subterms
		     subtype-replace intersection-type? find-intersection-subtyping find-intersection-supertyping
		     is-bottom? is-top? is-error? gen-error-mismatch gen-error-rule
		     is-dec-type-expr? dec-type-expr->env)
	    (let ((mgu (mgu-proc md)))
	      (letrec
		  ((recurse
		    (lambda (ts rs)
		      (cond
		       ((and (null? ts) (null? rs)) (list nullSubst))
		       ((null? ts) '())
		       ((null? rs) '())
		       (else (let ((u1 (mgu (car ts) (car rs))))
			       (if (null? u1)
				   '()
				   (let ((u2
					  (recurse (map (app (car u1)) (cdr ts))
						   (map (app (car u1)) (cdr rs)))  
					  ))
				     (if (null? u2)
					 '()
					 (list (tc:subst-compose
						(car u2) (car u1))))))))))))
		recurse))))))

(define tc:list-unify
  (lambda (md)
    (tc:list-unify-maker tc:mgu md)))

(define tc:list-subtype-unify
  (lambda (md)
    (tc:list-unify-maker tc:subtype-mgu md)))

(deftype remove-duplicates (forall (T) (-> ((list-of T)) (list-of T))))
(define remove-duplicates
  (lambda (l)
    (if (null? l)
	'()
	(cons (car l) (remove-all (car l) (remove-duplicates (cdr l)))))))

(deftype remove-all (forall (T) (-> (T (list-of T)) (list-of T))))
(define remove-all
  (lambda (x l)
    (cond
     ((null? l) '())
     ((equal? x (car l)) (remove-all x (cdr l)))
     (else (cons (car l) (remove-all x (cdr l)))))))
		    
;; Before trying to find the most general unifier of two terms,
;; one should make sure that they have no variables in common.
;; This can be accomplished by using the following function twice,
;; once for each term, but using two different integers.

(define tc:rename-vars
  (lambda (md)
    (cases tc:unifiable-md md
       (tc:unifiable-md-dict
	(to-term get-var subterms same-kind nullSubst bind app
		 contravar-subterms covar-subterms invar-subterms
		 subtype-replace intersection-type? find-intersection-subtyping find-intersection-supertyping
		 is-bottom? is-top? is-error? gen-error-mismatch gen-error-rule
		 is-dec-type-expr? dec-type-expr->env)
	     (lambda (lev)
	       (lambda (t)
		 ((app
		   (tc:all-subst to-term lev))
		  t)))))))

) ;; end module
